!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
module emission_common

  ! While the averagers could be put into each flux module, managing
  ! these routines is easier when all the repetitive code is gathered
  ! here in one place.

  use GO,                  only: gol, goErr, goPr, goLabel
  use Dims,                only: nlon360,nlat180

  public   :: co2_ocean, co2_fires, co2_ff, co2_bio, co2_all
  public   :: co2_ocean_m, co2_fires_m, co2_ff_m, co2_bio_m, co2_all_m
  public   :: co2_ocean_n, co2_fires_n, co2_ff_n, co2_bio_n, co2_all_n
  public   :: declare_emission_averagers,free_emission_averagers,reset_emission_averagers, calc_emission_averagers, save_emission_averagers
  public   :: enkf_is_inverse, enkf_is_forward, enkf_set_inverse, enkf_set_forward
  public   :: flux_means_are_priors, EmisInputDir

  character(len=*), parameter   ::  mname = 'emission_common'
  private:: mname

  real,dimension(nlon360,nlat180),target :: co2_ocean
  real,dimension(nlon360,nlat180),target :: co2_fires
  real,dimension(nlon360,nlat180),target :: co2_bio
  real,dimension(nlon360,nlat180),target :: co2_ff
  real,dimension(nlon360,nlat180),target :: co2_all

  ! weekly average flux accumulators.  "_m" is the flux field itself,
  ! "_n" is the number of accumulations.

  real,allocatable,dimension(:,:) :: co2_ocean_m
  real,allocatable,dimension(:,:) :: co2_fires_m
  real,allocatable,dimension(:,:) :: co2_bio_m
  real,allocatable,dimension(:,:) :: co2_ff_m
  real,allocatable,dimension(:,:) :: co2_all_m

  integer                                :: co2_ocean_n
  integer                                :: co2_fires_n
  integer                                :: co2_bio_n
  integer                                :: co2_ff_n
  integer                                :: co2_all_n

  ! Are we running in inverse or forward mode?  Set enkf mode in Emis_Init().
  integer,parameter             :: enkf_mode_unknown = 0
  integer,parameter             :: enkf_mode_inverse = 1
  integer,parameter             :: enkf_mode_forward = 2
  integer                       :: enkf_mode = enkf_mode_unknown

  character(len=256)            :: EmisInputDir

  ! Weekly mean fluxes can either be means of the raw priors, or means
  ! of the optimized fluxes.  Default is priors.  This was the
  ! behavior in previous versions, since one needs to verify what the
  ! flux modules are doing, and it is possible to retrieve the scaled
  ! fluxes in postprocessing.  This switch removes the need for that
  ! postprocessing, and may be useful in a future with a more complex
  ! parameter model.
  !
  ! flux_means_are_priors = .TRUE.  -> report weekly means of prior fluxes
  !
  ! flux_means_are_priors = .FALSE.  -> report weekly means of scaled/parameterized/optimized fluxes
  !
  logical                        :: flux_means_are_priors = .TRUE.


contains

    logical function enkf_is_inverse()
      enkf_is_inverse = .FALSE.
      if(enkf_mode .EQ. enkf_mode_inverse) then
         enkf_is_inverse = .TRUE.
      endif
    end function enkf_is_inverse

    logical function enkf_is_forward()
      enkf_is_forward = .FALSE.
      if(enkf_mode .EQ. enkf_mode_forward) then
         enkf_is_forward = .TRUE.
      endif
    end function enkf_is_forward

    ! Comment on flux_means_are_priors  (Andy Jacobson, 29 Sep 2007)
    !
    ! Wouter and I decided that the inverse run will output weekly-average
    ! prior fluxes by default, and that a follow-on forward run (which is 
    ! already pretty much a requirement for various output) will write out 
    ! the posterior fluxes.  Thus the default values for flux_means_as_priors
    ! below.  Note that in the inv and fwd versions of emission.F90, there
    ! is a commented-out ReadRC call that could be used to change this
    ! hard-coded default behavior.  Use of old RC files with flux_means_are_priors
    ! lines may cause some confusion, but the variables written to the 
    ! output file actually are given explicit names so it will be tough
    ! to make a real mistake.
    !
    ! It is convenient to look at posterior fluxes while an inverse run is
    ! underway.  For this, we have two suggestions.  First, the current model
    ! for fluxes is that the posterior is a scaled version of the prior.  The
    ! scale factors are the EnKF parameters, and are written to the savestate
    ! file along with the prior fluxes in the inverse run.  Thus, the weekly
    ! mean posterior fluxes can be constructed from those outputs.  This will
    ! not work if the parameter model changes such that we are optimizing
    ! parameters which affect the flux in a nonlinear fashion.  Second,
    ! the forward run is generally much, much cheaper to run and should be
    ! able to generate "real" posterior fluxes quite quickly.


    subroutine enkf_set_inverse
      enkf_mode=enkf_mode_inverse
      flux_means_are_priors=.TRUE.
      ! print *,'EnKF mode set to inverse'
    end subroutine enkf_set_inverse

    subroutine enkf_set_forward
      enkf_mode=enkf_mode_forward
      ! print *,'EnKF mode set to forward'
      flux_means_are_priors=.FALSE.
    end subroutine enkf_set_forward

  subroutine declare_emission_averagers
!    print *,"DEBUG: in declare_emission_averagers"

!WP! Always write component fluxes to savestate.hdf for analysis
    !WP!if(enkf_is_forward()) then
       allocate(co2_bio_m(nlon360,nlat180))
       allocate(co2_fires_m(nlon360,nlat180))
       allocate(co2_ocean_m(nlon360,nlat180))
       allocate(co2_ff_m(nlon360,nlat180))
    !WP!else
    !WP!  allocate(co2_all_m(nlon360,nlat180))
    !WP!endif

    call reset_emission_averagers

  end subroutine declare_emission_averagers

  subroutine free_emission_averagers

    if(allocated(co2_all_m)) then
       deallocate(co2_all_m)
    endif

    if(allocated(co2_ocean_m)) then
       deallocate(co2_ocean_m)
    endif

    if(allocated(co2_fires_m)) then
       deallocate(co2_fires_m)
    endif

    if(allocated(co2_bio_m)) then
       deallocate(co2_bio_m)
    endif

    if(allocated(co2_ff_m)) then
       deallocate(co2_ff_m)
    endif

  end subroutine free_emission_averagers

  subroutine calc_emission_averagers

    if(allocated(co2_all_m)) then       
       if(co2_all_n .GT. 0) then
         co2_all_m=co2_all_m/co2_all_n
      endif
    endif

    if(allocated(co2_ff_m)) then       
       if(co2_ff_n .GT. 0) then
         co2_ff_m=co2_ff_m/co2_ff_n
      endif
    endif

    if(allocated(co2_bio_m)) then       
       if(co2_bio_n .GT. 0) then
         co2_bio_m=co2_bio_m/co2_bio_n
      endif
    endif

    if(allocated(co2_fires_m)) then       
       if(co2_fires_n .GT. 0) then
         co2_fires_m=co2_fires_m/co2_fires_n
      endif
    endif

    if(allocated(co2_ocean_m)) then       
       if(co2_ocean_n .GT. 0) then
         co2_ocean_m=co2_ocean_m/co2_ocean_n
      endif
    endif

  end subroutine calc_emission_averagers

  subroutine save_emission_averagers(hdf_file)

    use io_hdf
    use file_hdf,             only: WriteAttribute, THdfFile
    use ParTools,             only: myid, root

    type(THdfFile), intent(in)         ::  hdf_file
    integer               :: status
    character(len=200)    :: dsname,attname,comp,pp
    character(len=*), parameter  ::  rname = mname//'/save_emission_averagers'
    
    if(myid .ne. root) return
    pp='post'
    if(flux_means_are_priors .eq. .True.) then
       pp='prior'
    endif

    if(allocated(co2_all_m)) then
       comp='all'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',co2_all_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, co2_all_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(co2_ff_m)) then
       comp='ff'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',co2_ff_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, co2_ff_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(co2_bio_m)) then
       comp='bio'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',co2_bio_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, co2_bio_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(co2_fires_m)) then
       comp='fires'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',co2_fires_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, co2_fires_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(co2_ocean_m)) then
       comp='ocean'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',co2_ocean_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, co2_ocean_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

  end subroutine save_emission_averagers

  subroutine reset_emission_averagers

    if(allocated(co2_all_m)) then
       co2_all_m=0.
       co2_all_n=0
    endif

    if(allocated(co2_ocean_m)) then
       co2_ocean_m=0.
       co2_ocean_n=0
    endif

    if(allocated(co2_fires_m)) then
       co2_fires_m=0.
       co2_fires_n=0
    endif

    if(allocated(co2_bio_m)) then
       co2_bio_m=0.
       co2_bio_n=0
    endif

    if(allocated(co2_ff_m)) then
       co2_ff_m=0.
       co2_ff_n=0
    endif

  end subroutine reset_emission_averagers

end module emission_common
