!#################################################################
!
! contains routines to read emissions and 
! to read and write the main model state from/to file
!
!### macro's #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
  

module io_save

  use GO, only : gol, goErr, goPr
  use GO, only : TrcFile, Init, Done, ReadRc
  use global_data,   only : rcfile
  
#include "advect.h"

  implicit none
  
  ! --- in/out --------------------------

  private

  public :: readhdf, readhdfmmr, savehdf, readhdfmmix
  public :: readtm3hdf, ReadFromFile


  ! --- const ----------------------------------

  character(len=*), parameter  ::  mname = 'io_save'
  type(TrcFile)       ::  rcF


contains


  ! ======================================================================

    subroutine readtm3hdf( filename, rank, im_emis, jm_emis, lm_emis, &
                         month, field, field_name, idt )
    use toolbox, only : escape_tm
    
    implicit none

    ! in/out
    character(len=*),intent(in)  :: filename  
    character(len=*),intent(in)  :: field_name
    integer,intent(in)           :: month 
    real,dimension(:,:),intent(out):: field
    integer,dimension(6),optional :: idt

    integer,intent(in)           :: rank
    integer,intent(in)           :: im_emis, jm_emis, lm_emis 
    integer                      :: status

    if (present(idt)) then
        call ReadFromFile(filename,field_name,field,status,sds_date=idt)
    else if (month /= 0 ) then
        call ReadFromFile(filename,field_name,field,status, sds_number=month+1 ) ! readtm3hdf used 0-indexing
    else 
        call ReadFromFile(filename,field_name,field,status ) ! readtm3hdf used 0-indexing
    endif
    if (status /= 0) call escape_tm("Stop: Error in readtm3hdf")

    end subroutine readtm3hdf

    subroutine ReadFromFile( filename, sds_name , data_array, status, sds_number, sds_date)
    use GO        , only : gol, goErr, goPr, goLabel, goUpCase
    use file_hdf

    !I/O
    character(len=*),intent(in)  :: filename  
    character(len=*),intent(in)  :: sds_name
    real,dimension(:,:),intent(out):: data_array
    integer, optional            :: sds_number
    integer,dimension(6),optional:: sds_date
    integer, intent (inout)      :: status
   

    type(THdfFile)      ::  hdf
    type(TSds)          ::  sds
    integer             ::  n, k
    character(len=*), parameter ::  rname = mname//'/ReadFromFile'

    ! start
    call Init( hdf, filename, 'read', status )
    IF_ERROR_RETURN(status=1)
    call Init( sds, status )
    IF_ERROR_RETURN(status=1)
    call GetInfo( hdf, status, num_datasets=n )
    IF_ERROR_RETURN(status=1)
    k = 0
    do
      k = k + 1
      if ( k > n ) then 
          write(gol,*) 'SDS not found' ; call goPr
          write(gol,*) 'filename   : ', trim(filename) ; call goPr
          write(gol,*) 'sds_name   : ', trim(sds_name) ; call goPr
          if (present(sds_number)) write(gol,*) 'sds_number :',sds_number ; call goPr
          if (present(sds_date))   write(gol,*) 'sds_date   :',sds_date ; call goPr
          IF_NOTOK_RETURN(status=1)
      endif

      call Select( sds, hdf, k-1, status )     ! <-- sets numbered 0..n-1 !
      status=1
      call CheckInfo( sds, status, name=sds_name )

      if ( status /= 0 ) cycle  ! name not correct

      status=1  ! for non-verbose checking of date attribute

      if (present(sds_date))   call CheckAttribute( sds, 'idate', sds_date , status )
      if (present(sds_number).and.k == sds_number) status=0
      if (present(sds_number) == 0.and. present(sds_date) == 0) status=0   !  automatic pass if no keywords

      if ( status /= 0 ) cycle  ! date not found, or sds_number not reached

      ! all lights green, go read

      call ReadData( sds, data_array, status )
      IF_ERROR_RETURN(status=1)
      exit ! break from loop

    end do
    call Done( sds, status )
    IF_ERROR_RETURN(status=1)
    call Done( hdf, status )
    IF_ERROR_RETURN(status=1)
    status=0

    end subroutine ReadFromFile


  subroutine readtm3hdf_old( filename, rank, im_emis, jm_emis, lm_emis, &
                         month, field, field_name, idt )
    use dims
    use toolbox, only : escape_tm
    use io_hdf,  only : DFACC_READ, FAIL
    use io_hdf,  only : io_read2d_32g, io_read3d_32g, io_read1d_32g
#ifdef MPI
     use mpi_const,   only : myid
#endif     
    
    implicit none

    !WP! this subroutine reads HDF emissions files on 
    !    1x1 degree resolution and returns the specified field.
    !WP! input for this subroutine are:
    !WP!
    !WP! filename --> name of hdf file to be opened or read
    !MK! rank    --> rank of the dataset (1,2,3)
    !WP! im_emis --> x-DIMENSION of requested field
    !WP! jm_emis --> y-DIMENSION of requested field     
    !WP! lm_emis --> z-DIMENSION of requested field
    !WP! month  --> the month to read (0 for january or simply the first field)
    !WP! field --> the field (im_emis,jm_emis,lm_emis) to put the values in
    !WP! field_name --> the name of the field to match the one in the HDF file
    !WP!
    !WP! The routine checks the dimensions and reports the succes/failure.

    ! in/out
    character(len=*),intent(in)  :: filename
    character(len=*),intent(in)  :: field_name
    integer,intent(in)           :: rank
    integer,intent(in)           :: month
    integer,intent(in)           :: im_emis, jm_emis, lm_emis
    real,dimension(im_emis,jm_emis,lm_emis),intent(out)  :: field
    integer,dimension(6),optional :: idt

    ! local
    real,dimension(:),allocatable       :: field1d   ! first dim will be 1
    real,dimension(:,:),allocatable     :: field2d
    real,dimension(:,:,:),allocatable   :: field3d

    integer :: io, sfstart, ifail, sds_id, ioo
    integer :: n_datasets, n_file_attrs
    integer :: istat, attributes, num_type
    integer :: sffinfo, sfselect, sfginfo
    integer :: sfend, sfrnatt, sfrcatt, sffattr
    integer :: idx,l
    integer :: region,ntot

    ! start

    idx=month    !FD hdf files indices run from 0 ...n-1
    io = SFSTART(trim(filename), DFACC_READ)  !WP! assign io-number to HDF file
    
    if ( io /= FAIL ) then
      ! print *,'readtm3hdf: ',filename,'... opened for READ access.',io 
    else
      ! print *,'readtm3hdf: File name:',filename
       call escape_tm('readtm3hdf: Unable to open HDF file '//&
            trim(filename)//' in READTM3HDF ')
    end if

    select case(rank) 
    case(1)    ! latitudional field

       allocate(field1d(jm_emis))
       if(present(idt)) call io_read1d_32g(io,jm_emis,field1d,field_name, ifail,idate=idt)
       if(.not.present(idt)) call io_read1d_32g(io,jm_emis,field1d,field_name, ifail,index=idx)
       if ( ifail /= 0 ) then
          print *, 'readtm3hdf: io_read1d_32g failed:'
          print *, 'readtm3hdf:   filename   : ', trim(filename)
          print *, 'readtm3hdf:   rank       : ', rank
          print *, 'readtm3hdf:   jm_emis    : ', jm_emis
          print *, 'readtm3hdf:   month      : ', month
          print *, 'readtm3hdf:   field_name : ', trim(field_name)
          if (present(idt)) print *, 'readtm3hdf:   idt        : ', idt
          call escape_tm('readtm3hdf: Error reading HDF file 1D')
       end if
       field(1,:,1)=field1d
       deallocate(field1d)

    case(2)    ! 2D surface field

       allocate(field2d(im_emis,jm_emis))
       if(present(idt)) call io_read2d_32g(io,im_emis,jm_emis,field2d,field_name, ifail,idate=idt)
       if(.not.present(idt)) call io_read2d_32g(io,im_emis,jm_emis,field2d,field_name, ifail,index=idx)
       if ( ifail /= 0 ) then
          print *, 'readtm3hdf: io_read2d_32g failed:'
          print *, 'readtm3hdf:   filename   : ', trim(filename)
          print *, 'readtm3hdf:   rank       : ', rank
          print *, 'readtm3hdf:   im_emis    : ', im_emis
          print *, 'readtm3hdf:   jm_emis    : ', jm_emis
          print *, 'readtm3hdf:   lm_emis    : ', lm_emis
          print *, 'readtm3hdf:   month      : ', month
          print *, 'readtm3hdf:   field_name : ', trim(field_name)
          if (present(idt)) print *, 'readtm3hdf:   idt        : ', idt
          call escape_tm('readtm3hdf: Error reading HDF file 2D')
       endif
       field(:,:,1)=field2d
       deallocate(field2d)

    case (3)   ! 3D field
     
       allocate(field3d(im_emis,jm_emis,lm_emis))
       if(present(idt)) call io_read3d_32g(io,im_emis,jm_emis,lm_emis,field3d,field_name, ifail,idate=idt)
       if(.not.present(idt)) call io_read3d_32g(io,im_emis,jm_emis,lm_emis,field3d,field_name, ifail,index=idx)

       if ( ifail /= 0 )  then
          print *, 'readtm3hdf: io_read2d_32g failed:'
          print *, 'readtm3hdf:   filename   : ', trim(filename)
          print *, 'readtm3hdf:   rank       : ', rank
          print *, 'readtm3hdf:   im_emis    : ', im_emis
          print *, 'readtm3hdf:   jm_emis    : ', jm_emis
          print *, 'readtm3hdf:   lm_emis    : ', lm_emis
          print *, 'readtm3hdf:   month      : ', month
          print *, 'readtm3hdf:   field_name : ', trim(field_name)
          if (present(idt)) print *, 'readtm3hdf:   idt        : ', idt
          call escape_tm('readtm3hdf: Error reading HDF file 3D')
       endif
       field(:,:,:)=field3d
   
       deallocate(field3d)

    case (23)   ! 2D field lat-pres

       allocate(field2d(jm_emis,lm_emis))
       if(present(idt)) call io_read2d_32g(io,jm_emis,lm_emis,field2d,field_name, ifail,idate=idt)
       if(.not.present(idt)) call io_read2d_32g(io,jm_emis,lm_emis,field2d,field_name, ifail,index=idx)
       if ( ifail /= 0 )  then   !might be due to (1,jm,lm) file!
           print *, 'readtm3hdf: try shape : ', 1, jm_emis, lm_emis
          allocate(field3d(1,jm_emis,lm_emis))
          if ( present(idt) ) then
            call io_read3d_32g(io,1,jm_emis,lm_emis,field3d,field_name, ifail,idate=idt)
          else
            call io_read3d_32g(io,1,jm_emis,lm_emis,field3d,field_name, ifail,index=idx)
          end if
          if ( ifail /= 0 ) then
             print *, 'readtm3hdf: io_read2d_32g failed:'
             print *, 'readtm3hdf:   filename   : ', trim(filename)
             print *, 'readtm3hdf:   rank       : ', rank
             print *, 'readtm3hdf:   im_emis    : ', im_emis
             print *, 'readtm3hdf:   jm_emis    : ', jm_emis
             print *, 'readtm3hdf:   lm_emis    : ', lm_emis
             print *, 'readtm3hdf:   month      : ', month
             print *, 'readtm3hdf:   field_name : ', trim(field_name)
             if (present(idt)) print *, 'readtm3hdf:   idt        : ', idt
             call escape_tm('readtm3hdf: Error reading HDF file 23')
          end if
          field(:,:,:)=field3d
          deallocate(field3d)
       else
          field(1,:,:)=field2d
          deallocate(field2d)
       end if

    case default
       print *,'readtm3hdf: Please check the rank of the required field,', &
               ' must be 1,2,3, or 23'
    end select

    istat = SFEND(io)  !close hdf file
    !if ( idx == 0 ) print *,'readtm3hdf: ',trim(filename), &
    !     ' total sum after reading the full field',sum(field)/1e9 ,' (x1e9) '
    !if ( idx > 0 ) print *,'readtm3hdf: ',trim(filename),' ds#= ',idx+1, &
    !     ' total month sum after reading the full field', &
    !     sum(field)/1e9 ,' (x1e9) '

  end subroutine readtm3hdf_old


  ! ***
  

  !
  ! data are expected in kg   
  ! (typical save.hdf files from previous run...)
  ! only long-lived tracer are read, since the fields 
  ! for chemistry are not yet allocated
  !

  subroutine ReadHDF( region, file_name )

    use io_hdf
    use dims,        only : region_name, im, jm, lm, nregions, &
                            datadir, parent, adv_scheme
    use global_data, only : mass_dat
    use chem_param,  only : ntrace, ntracet
    use toolbox,     only : escape_tm
    use ParTools   , only : myid, root_t
    use ParTools   , only : ntracet_ar, ntracetloc
#ifdef MPI
    use mpi_comm,    only : scatter_after_read_t
    use mpi_const,   only : com_trac, ierr, my_real, mpi_integer
#endif

    ! --- in/out ------------------------------

    integer, intent(in)             :: region
    character(len=*), intent(in)    :: file_name 
    character(len=40)  :: xx
    
    ! --- local -------------------------------

    real,dimension(:,:,:,:),pointer :: rm,rxm,rym,rzm
#ifdef secmom
    real,dimension(:,:,:,:),pointer :: rxxm,rxym,rxzm,ryym,ryzm,rzzm
#endif
    real,dimension(:,:,:),pointer   :: m

    integer                         :: io, sfstart, ifail, sds_id
    integer                         :: n_datasets, n_file_attrs
    integer                         :: istat, attributes, num_type
    integer                         :: sffinfo, sfselect, sfginfo
    integer                         :: sfend, sfrnatt, sfrcatt, sffattr
    integer,dimension(6)            :: idate_save
    integer,dimension(nregions)     :: im_file,jm_file,lm_file
    integer                         :: ntrace_file, i,j,l,n
    character(len=80)               :: msg_file
    character(len=256)              :: fname
    integer                         :: ind, my_parent,imr,jmr,lmr
    real,dimension(:,:,:,:), allocatable :: field, field2
    real,dimension(:,:,:), allocatable   :: msave
    integer                         :: from_file  ! signals reading from file

    ! --- begin --------------------------------

    write(xx,*) myid
    xx=adjustl(xx)
    fname=trim(file_name)//'_pe'//trim(xx)

    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
#ifdef secmom
    rxxm => mass_dat(region)%rxxm_t
    rxym => mass_dat(region)%rxym_t
    rxzm => mass_dat(region)%rxzm_t
    ryym => mass_dat(region)%ryym_t
    ryzm => mass_dat(region)%ryzm_t
    rzzm => mass_dat(region)%rzzm_t
#endif

    imr = im(region) ; jmr = jm(region) ; lmr = lm(region)

    allocate(field(-1:imr+2,-1:jmr+2,lmr,ntracetloc))
    field = 0.0

    allocate(msave(imr,jmr,lmr))

    from_file = 1
    allocate(field2(imr,jmr,lmr,ntracetloc))   
    field2 = 0.0

    ind = len_trim(datadir)

    io = sfstart( trim(fname), DFACC_READ )
    if ( io == FAIL ) then 
      my_parent = parent(region)    
      if ( my_parent == 0 ) then
         call escape_tm('readhdf: Could not open file and no parent '// &
              'available :'//trim(fname) )
      else
         from_file = 0
         write (gol,'("readhdf: Trying to initialise ",a," from parent....")') region_name(region) ; call goPr
      end if
    end if
    if ( from_file == 1 ) then
      write (gol,'("readhdf:")'); call goPr
      write (gol,*)"readhdf: opened for READ access: ",trim(fname); call goPr

      call io_read3d_32g(io,imr,jmr,lmr,msave,'m',ifail)
      if ( ifail /= 0 ) then
         call escape_tm('readhdf: Failed to read m from saveold.hdf in readhdf')
      end if

      call io_read4d_32g(io,imr,jmr,lmr,ntracetloc,field2,'rm',ifail)
      if( ifail /= 0 .and. ntracetloc == 1 ) then
         call io_read3d_32g(io,imr,jmr,lmr,field2,'rm',ifail)
      end if

      if ( ifail /= 0 ) &
           call escape_tm('readhdf: Failed to read in saveold.hdf in readhdf')
      write (gol,*)"readhdf: Read rm, ifail = ",ifail; call goPr

      field(1:imr,1:jmr,1:lmr,:)=field2 ! only transported tracers!

      rm = field
    else  ! from_file
       call init_child(region)   ! get info from parent....
    end if

    field = 0.0

    if ( from_file == 1 ) then   
       if ( adv_scheme == 'slope'.or.adv_scheme.eq.'2nd_m' ) then 

             call io_read4d_32g(io,imr,jmr,lmr,ntracetloc,field2,'rxm',ifail)
             if ( ifail /= 0 .and. ntracetloc == 1 ) then
                call io_read3d_32g(io,imr,jmr,lmr,field2,'rxm',ifail)
             end if
             if ( ifail /= 0 ) then
                call escape_tm( &
                  'readhdf: Failed to read in rxm in saveold.hdf in readhdf')
             end if
      	     write (gol,*)"readhdf: Read rxm, ifail = ",ifail; call goPr

             field(1:imr,1:jmr,1:lmr,:)=field2 ! only transported tracers!

             rxm = field
             field = 0.0

             call io_read4d_32g(io,imr,jmr,lmr,ntracetloc,field2,'rym',ifail)
             if ( ifail /= 0 .and. ntracetloc == 1 ) then
                call io_read3d_32g(io,imr,jmr,lmr,field2,'rym',ifail)
             end if
             if ( ifail /= 0 ) then
                call escape_tm( &
                     'readhdf: Failed to read in rym in saveold.hdf in readhdf')
             end if
      	     write (gol,*)"readhdf: Read rym, ifail = ",ifail; call goPr
             field(1:imr,1:jmr,1:lmr,:)=field2 ! only transported tracers!

             rym = field
             field = 0.0

             call io_read4d_32g(io,imr,jmr,lmr,ntracetloc,field2,'rzm',ifail)
             if ( ifail /= 0 .and. ntracetloc == 1 ) then
                call io_read3d_32g(io,imr,jmr,lmr,field2,'rzm',ifail)
             end if
             if ( ifail /= 0 ) then
                call escape_tm( &
                     'readhdf: Failed to read in rzm in saveold.hdf in readhdf')
             end if
      	     write (gol,*)"readhdf: Read rzm, ifail = ",ifail; call goPr

             field(1:imr,1:jmr,1:lmr,:)=field2 ! only transported tracers!

             rzm = field

             deallocate(field2)

       end if  !slope

       istat = sfend(io)

       if ( istat /= FAIL ) then
          write (gol,'("readhdf: ... file closed")'); call goPr
          write (gol,'("readhdf: ")'); call goPr
       else
          call escape_tm('readhdf: ERROR in restart from HDF file')
       end if

   ! scale in input by a possible factor.
   ! rm/msave ---> mixing ratio on file. mr*m --> something back in kilo's
       if(myid==root_t) then
           write(gol,*) "readhdf: Maximum ratio of the air masses:", &
                maxval(abs(msave/m(1:imr,1:jmr,1:lmr))); call goPr
           write(gol,*) "readhdf: Minimum ratio of the air masses:", &
                minval(abs(msave/m(1:imr,1:jmr,1:lmr))); call goPr
       endif
       do n=1,ntracetloc
          do l = 1,lmr
             do j = 1,jmr
                do i = 1,imr
                   rm(i,j,l,n)  = rm (i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   rxm(i,j,l,n) = rxm(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   rym(i,j,l,n) = rym(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   rzm(i,j,l,n) = rzm(i,j,l,n)*m(i,j,l)/msave(i,j,l)
#ifdef secmom
                   rxxm(i,j,l,n) = rxxm(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   rxym(i,j,l,n) = rxym(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   rxzm(i,j,l,n) = rxzm(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   ryym(i,j,l,n) = ryym(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   ryzm(i,j,l,n) = ryzm(i,j,l,n)*m(i,j,l)/msave(i,j,l)
                   rzzm(i,j,l,n) = rzzm(i,j,l,n)*m(i,j,l)/msave(i,j,l)
#endif
                end do
             end do
          end do
       end do
       
    end if   ! from_file == 1

    deallocate(msave)
    deallocate(field)
    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
#ifdef secmom
    nullify(rxxm)
    nullify(rxym)
    nullify(rxzm)
    nullify(ryym)
    nullify(ryzm)
    nullify(rzzm)
#endif

  end subroutine ReadHDF


  ! ***
  

  subroutine init_child(region)
    !
    ! Parent is used to initilise rm, rxm,rym,rzm of region. 
    ! Assumed parallel over tracers...
    !
    use dims,        only : lm, ibeg, iend, jbeg, jend, xref, yref, parent
    use global_data, only : mass_dat
    use chem_param,  only : ntrace, ntracet
    use toolbox,     only : escape_tm
    use ParTools,    only : myid, ntracetloc, root_t

    implicit none

    ! in/out
    integer, intent(in) :: region

    ! local
    real,dimension(:,:,:,:),pointer :: rm,rxm,rym,rzm   ! child
    real,dimension(:,:,:),pointer   :: m
    real,dimension(:,:,:,:),pointer :: rmp   !parent
    real,dimension(:,:,:),pointer   :: mp

    real           :: mm_parent
    integer        :: my_parent
    integer        :: n,l,jp,ip,ic,jc, xref_, yref_, i,j

    ! start

    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t

    my_parent = parent(region)    

    if ( my_parent == 0 ) then
       call escape_tm( 'init_child: Region 1 cannot be initialised')
    end if

    mp => mass_dat(my_parent)%m_t
    rmp => mass_dat(my_parent)%rm_t

    xref_ = xref(region)/xref(my_parent)
    yref_ = yref(region)/yref(my_parent)
    do n=1,ntracetloc
       do l=1,lm(region)
          do jp=jbeg(region), jend(region)
             jc = (jp-jbeg(region))*yref_
             do ip=ibeg(region), iend(region)
                mm_parent = rmp(ip,jp,l,n)/mp(ip,jp,l)
                ic = (ip-ibeg(region))*xref_
                do j=1,yref_
                   do i=1,xref_
                      rm (ic+i,jc+j,l,n) = mm_parent*m(ic+i,jc+j,l)  
                      rxm(ic+i,jc+j,l,n) = 0.0
                      rym(ic+i,jc+j,l,n) = 0.0
                      rzm(ic+i,jc+j,l,n) = 0.0
                   end do
                end do
             end do
          end do
       end do
    end do
    if ( myid == root_t ) print *, 'readhdf: Initialisation successful '

    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(mp)
    nullify(rmp)
    
  end subroutine init_child

  subroutine init_child_edge(region)
    !
    ! ONLY the edges!
    ! Parent is used to initilise rm, rxm,rym,rzm of region. 
    ! Assumed parallel over tracers...
    !
    use dims,        only : lm, ibeg, iend, jbeg, jend, xref, yref, parent, im, jm
    use global_data, only : mass_dat, region_dat
    use chem_param,  only : ntrace, ntracet
    use toolbox,     only : escape_tm
    use ParTools,    only : myid, ntracetloc, root_t

    implicit none

    ! in/out
    integer, intent(in) :: region

    ! local
    real,dimension(:,:,:,:),pointer :: rm,rxm,rym,rzm   ! child
    real,dimension(:,:,:),pointer   :: m
    real,dimension(:,:,:,:),pointer :: rmp   !parent
    real,dimension(:,:,:),pointer   :: mp
    integer,dimension(:,:),pointer  :: edge

    real           :: mm_parent
    integer        :: my_parent
    integer        :: n,l,jp,ip,ic,jc, xref_, yref_, i,j

    ! start

    my_parent = parent(region)    
    if ( my_parent == 0 ) return

    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
    edge => region_dat(region)%edge


    mp => mass_dat(my_parent)%m_t
    rmp => mass_dat(my_parent)%rm_t

    xref_ = xref(region)/xref(my_parent)
    yref_ = yref(region)/yref(my_parent)
    do n=1,ntracetloc
       do l=1,lm(region)
          do j=1,jm(region)
             jp = jbeg(region) + (j-1)/yref_
             do i=1,im(region)
                if (edge(i,j) /= -1 ) cycle
                ip = ibeg(region) + (i-1)/xref_
                mm_parent = rmp(ip,jp,l,n)/mp(ip,jp,l)
                rm (i,j,l,n) = mm_parent*m(i,j,l)  
                rxm(i,j,l,n) = 0.0
                rym(i,j,l,n) = 0.0
                rzm(i,j,l,n) = 0.0
             end do
          end do
       end do
    end do
    if ( myid == root_t ) print *, 'readhdf: Edges Initialisation successful '

    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(mp)
    nullify(rmp)
    
  end subroutine init_child_edge





  subroutine readhdfmmr(region,file_name)
    !
    ! data are expected in mixing ratio...
    !
    use dims,        only : im, jm, lm, nregions, datadir, region_name
    use io_hdf
    use global_data, only : mass_dat
    use chem_param,  only : fscale, ntracet
    use toolbox,     only : escape_tm
#ifdef MPI
    use mpi_comm,    only : scatter_after_read_t
    use mpi_const,   only : root, myid, ntracet_ar, ntracetloc
#endif

    implicit none

    ! in/out
    integer, intent(in)             :: region
    character(len=*), intent(in)    :: file_name 

    ! local
    real,dimension(:,:,:,:),pointer :: rm,rxm,rym,rzm
#ifdef secmom
    real,dimension(:,:,:,:),pointer :: rxxm,rxym,rxzm,ryym,ryzm,rzzm
#endif
    real,dimension(:,:,:),pointer   :: m
    integer                         :: io, sfstart, ifail, sds_id
    integer                         :: n_datasets, n_file_attrs
    integer                         :: istat, attributes, num_type
    integer                         :: sffinfo, sfselect, sfginfo
    integer                         :: sfend, sfrnatt, sfrcatt, sffattr
    integer,dimension(6)            :: idate_save
    integer,dimension(nregions)     :: im_file,jm_file,lm_file
    integer                         :: ntrace_file
    character(len=80)               :: msg_file
    integer                         :: ind,ind1,n,i,j,l,offsetn
    real,dimension(:,:,:,:), allocatable :: field,field2

    ! start

    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
#ifdef secmom
    rxxm => mass_dat(region)%rxxm_t
    rxym => mass_dat(region)%rxym_t
    rxzm => mass_dat(region)%rxzm_t
    ryym => mass_dat(region)%ryym_t
    ryzm => mass_dat(region)%ryzm_t
    rzzm => mass_dat(region)%rzzm_t
#endif
    allocate(field(-1:im(region)+2,-1:jm(region)+2,1:lm(region),1:ntracet))
    field=0.0
#ifdef MPI
    if ( myid == root ) then 
#endif
       ! to hold all ntrace values of rm
       allocate(field2(im(region),jm(region),lm(region),ntracet))
       io = sfstart( file_name, DFACC_READ )

       if ( io == -1 ) call escape_tm('readhdfmmr: Could not open file:'//file_name)
       ind = 1  ! select 1st dataset....(0 = m)

       istat = sffinfo(io, n_datasets, n_file_attrs)
       print*, 'readhdfmmr: Number of datasets and attributes',  &
            n_datasets, n_file_attrs
       call io_read4d_32g(io,im(region),jm(region),lm(region),ntracet, &
            field2,'rm',ifail,index=ind)
       if ( ifail /= 0 .and. ntracet == 1 ) then
          call io_read3d_32g(io,im(region),jm(region),lm(region), &
               field2,'rm',ifail,index=ind)
       end if
       if ( ifail /= 0 ) then
          print *, 'readhdfmmr: '//datadir(1:ind)//file_name(1:ind1-1)// &
            '_'//region_name(region)//'.hdf'
          call escape_tm('readhdfmmr: Failed to read fields')
       end if
       print*, 'readhdfmmr: Read rm, ifail = ', ifail

       istat = sfend(io)
       if (istat /= FAIL) then
          print*,'readhdfmmr: ... file closed'
          print*,' '
       else
          call escape_tm('readhdfmmr: ERROR in restart from HDF file')
       end if

       ! only transported tracers!
       field(1:im(region),1:jm(region),1:lm(region),:)=field2
       deallocate(field2) ! no longer needed
#ifdef MPI
    end if ! root
    call scatter_after_read_t(field,im(region),jm(region),lm(region),&
         2,2,0,rm,root)
#else
    rm = field
#endif
#ifdef MPI
    offsetn = sum(ntracet_ar(0:myid-1))
    do n=1,ntracetloc
#else
    offsetn = 0
    do n=1,ntracet
#endif
       do l = 1,lm(region)
          do j = 1,jm(region)
             do i = 1,im(region)
                rm(i,j,l,n) = rm(i,j,l,n)*m(i,j,l)/fscale(offsetn+n)
                rxm(i,j,l,n) = 0.0
                rym(i,j,l,n) = 0.0
                rzm(i,j,l,n) = 0.0
#ifdef secmom
                rxxm(i,j,l,n) = 0.0
                rxym(i,j,l,n) = 0.0
                rxzm(i,j,l,n) = 0.0
                ryym(i,j,l,n) = 0.0
                ryzm(i,j,l,n) = 0.0
                rzzm(i,j,l,n) = 0.0
#endif
             end do
          end do
       end do
    end do

    deallocate(field)

    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
#ifdef secmom
    nullify(rxxm)
    nullify(rxym)
    nullify(rxzm)
    nullify(ryym)
    nullify(ryzm)
    nullify(rzzm)
#endif
    !WP! checked rm with the values in savehdf after scatter 
    !    and gather on 4 processors=>OK

  end subroutine readhdfmmr

  subroutine readhdfmmix(region,file_name)
    !
    ! data are expected in mixing ratio from an mmix file
    !
    use dims,        only : im, jm, lm, nregions, datadir, region_name, parent
    use io_hdf
    use global_data, only : mass_dat
    use chem_param,  only : fscale, ntracet, names
    use toolbox,     only : escape_tm
#ifdef MPI
    use mpi_comm,    only : scatter_after_read_t
    use mpi_const,   only : ntracet_ar, ntracetloc
    use mpi_const,   only : com_trac, mpi_integer, root_t, ierr
#endif
    use ParTools,    only : root, myid, root_t

    implicit none

    ! in/out
    integer, intent(in)             :: region
    character(len=*), intent(in)    :: file_name 

    ! local
    real,dimension(:,:,:,:),pointer :: rm,rxm,rym,rzm
    real,dimension(:,:,:),pointer   :: m
    integer                         :: io, sfstart, ifail, sds_id
    integer                         :: n_datasets, n_file_attrs
    integer                         :: istat, attributes, num_type
    integer                         :: sffinfo, sfselect, sfginfo
    integer                         :: sfend, sfrnatt, sfrcatt, sffattr
    integer,dimension(6)            :: idate_save
    integer,dimension(nregions)     :: im_file,jm_file,lm_file
    integer                         :: ntrace_file
    character(len=80)               :: msg_file
    integer                         :: ind,ind1,n,i,j,l,offsetn, from_file, my_parent
    real,dimension(:,:,:,:), allocatable :: field
    real,dimension(:,:,:),   allocatable :: field3d

    ! start

    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
    rm = 0.0
    rxm = 0.0
    rym = 0.0
    rzm = 0.0
    allocate(field(-1:im(region)+2,-1:jm(region)+2,1:lm(region),1:ntracet))
    field = 0.0    !09/2004 cmk: to initialise the edges!
    if ( myid == root_t ) then    
       from_file = 1
       allocate(field3d(im(region),jm(region),lm(region)))
       io = sfstart(file_name, DFACC_READ)
       if ( io == FAIL ) then 
          my_parent = parent(region)    
          if ( my_parent == 0 ) then
             call escape_tm('readhdfmmix : Could not open file and no parent '// &
                  'available :'//file_name )
          else
             from_file = 0
             print *, 'readhdf: Trying to initialise '//region_name(region)// &
                  ' from parent....'
          endif
       endif   
       if ( from_file == 1 ) then   ! read from file:
          print*,' '
          print*,'readhdfmmix: ',file_name,'... opened for READ access.' 

       istat = sffinfo(io, n_datasets, n_file_attrs)

       print*, 'readhdfmmix: Number of datasets and attributes',  &
            n_datasets, n_file_attrs

       do n = 1, ntracet
          call io_read3d_32(io, im(region), jm(region), lm(region), field3d, names(n), ifail)
          if (ifail == 0) then
             field(1:im(region), 1:jm(region), 1:lm(region), n) = field3d
          else
             print *, 'Readhdfmmix:', names(n), '  not found in dataset --> 0.0 '
             field(1:im(region), 1:jm(region), 1:lm(region), n) = 0.0
          endif
       enddo   

       istat = sfend(io)
       if (istat /= FAIL) then
          print*,'readhdfmmix: ... file closed'
          print*,' '
       else
          call escape_tm('readhdfmmix: ERROR in restart from HDF file')
       end if

       deallocate(field3d) ! no longer needed
       endif  ! from file
    endif ! root_t   
             
#ifdef MPI
    ! broadcast from_file !
    call mpi_bcast(from_file, 1, mpi_integer , root_t, com_trac, ierr)
#endif
    if ( from_file == 1 ) then 
#ifdef MPI
      call scatter_after_read_t(field,im(region),jm(region),lm(region),&
           2,2,0,rm,root)
#else
      rm = field
#endif
#ifdef MPI
      offsetn = sum(ntracet_ar(0:myid-1))
      do n=1,ntracetloc
#else
      offsetn = 0
      do n=1,ntracet
#endif
         do l = 1,lm(region)
            do j = 1,jm(region)
               do i = 1,im(region)
                  rm(i,j,l,n) = rm(i,j,l,n)*m(i,j,l)/fscale(offsetn+n)  ! from mmix to kg tracer
                  rxm(i,j,l,n) = 0.0
                  rym(i,j,l,n) = 0.0
                  rzm(i,j,l,n) = 0.0
               end do
            end do
         end do
      end do

      deallocate(field)
      call init_child_edge(region)   ! get edge info from parent....
    else   ! initialise from parent:
       call init_child(region)   ! get info from parent....
    end if

    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    !WP! checked rm with the values in savehdf after scatter 
    !    and gather on 4 processors=>OK

  end subroutine readhdfmmix



  !------------------------------------------------------------------
  ! save all essential model parameters and fields on unit kdisk
  !
  !                                       v 8.4
  !  modified for HDF output
  !------------------------------------------------------------------

  subroutine SaveHDF( msg, file_name, status )

    use file_hdf   , only : THdfFile, Init, Done, WriteAttribute, TSds, SetDim, WriteData
    use dims       , only : splitsave
    use dims       , only : nregions, region_name
    use dims       , only : im, jm, lm
    use dims       , only : dx, xref, xbeg, xend, ibeg, iend
    use dims       , only : dy, yref, ybeg, yend, jbeg, jend
    use dims       , only : dz, zref, zbeg, zend, lbeg, lend
    use dims       , only : areag
    use dims       , only : czeta, czetak
    use dims       , only : itau, itaui, itaue, itau0, itaut
    use dims       , only : idate, idatei, idatee, idate0, idatet
    use dims       , only : icalendo, iyear0, julday0
    use dims       , only : newyr, newmonth, newday, newsrun
    use dims       , only : tref
    use dims       , only : ndyn, nconv, ndiag, nchem, nsrce
    use dims       , only : nread, ninst, ncheck, ndiff, ndiagp1, ndiagp2, nstep
    use dims       , only : nwrite
    use dims       , only : istart
    use dims       , only : cpu0, cpu1
    use dims       , only : xlabel
    use dims       , only : cdebug, kmain, kdebug
    use dims       , only : limits, adv_scheme
    use dims       , only : at, bt
    use dims       , only : nsplitsteps, splitorder
    use global_data, only : mass_dat
    use chem_param , only : ntrace, ntracet, ra, names, fscale
    use chem_param , only : nstd, istd
    use io_hdf,      only : io_write
    use datetime,    only : tstamp
    use toolbox,     only : escape_tm
    use partools   , only : myid, root_t, ntracetloc
    use partools   , only : Par_Barrier, Par_Gather_Tracer_t
    use partools   , only : which_par, previous_par

#ifdef pgf90
    use file_hdf_l
    use file_hdf_s
    use file_hdf_i1
    use file_hdf_i2
    use file_hdf_i4
    use file_hdf_i8
    use file_hdf_r4
    use file_hdf_r8
#endif
    
    ! --- in/out ----------------------------------

    character(len=*), intent(in)  ::  file_name
    character(len=*), intent(in)  ::  msg
    integer, intent(out)          ::  status
    
    ! --- const ----------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/SaveHDF'
    
    ! --- local ----------------------------------

    real, dimension(:,:,:,:), pointer       ::  rm,rxm,rym,rzm
#ifdef secmom
    real, dimension(:,:,:,:), pointer       ::  rxxm,rxym,rxzm,ryym,ryzm,rzzm
#endif
    real, dimension(:,:,:), pointer         ::  m
    real, dimension(:,:,:,:), allocatable   ::  fieldglob4d
    
    type(THdfFile)    ::  hdf
    type(TSds)        ::  sds

!    integer           :: istat, sfsnatt, sfscatt, sfstart
!    integer           :: sfcreate, sfscompress, sds_id, sfwdata
!    integer           :: sfendacc, sfend, dimid3, dimid0, dimid1
!    integer           :: dimid2, sfdimid, sfsdmname
    integer           :: region, imr,jmr, lmr,i

    character(len=64)   ::  msg_file
    character(len=256)  ::  fname
    character(len=40)  :: xx
    
    ! --- begin -----------------------------------

    if(ntracetloc==0) return

    if ( .not. splitsave ) call escape_tm('savehdf: splitsave should be true ')

    write(xx,*) myid
    xx=adjustl(xx)

    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)
    do region = 1,nregions

        !WP! read filename from rcfile
        call ReadRc(rcF, 'start.hdf.3.'//region_name(region), fname, status )
        IF_NOTOK_RETURN(status=1)

        fname=trim(fname)//'_pe'//trim(xx)

        call Init( hdf, trim(fname), 'create', status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'itau'       , itau               , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nregions'   , nregions           , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'region_name', region_name(region), status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'im'         , im(region)         , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'jm'         , jm(region)         , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'lm'         , lm(region)         , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'dx'         , dx/xref(region)    , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'dy'         , dy/yref(region)    , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'dz'         , dz/zref(region)    , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'xbeg'       , xbeg(region)       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'xend'       , xend(region)       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ybeg'       , ybeg(region)       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'yend'       , yend(region)       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'zbeg'       , zbeg(region)       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'zend'       , zend(region)       , status )
        IF_ERROR_RETURN(status=1)

        if ( region /= 1 ) then
          call WriteAttribute( hdf, 'ibeg', ibeg(region), status )
          IF_ERROR_RETURN(status=1)
          call WriteAttribute( hdf, 'iend', iend(region), status )
          IF_ERROR_RETURN(status=1)
          call WriteAttribute( hdf, 'jbeg', jbeg(region), status )
          IF_ERROR_RETURN(status=1)
          call WriteAttribute( hdf, 'jend', jend(region), status )
          IF_ERROR_RETURN(status=1)
          call WriteAttribute( hdf, 'lbeg', lbeg(region), status )
          IF_ERROR_RETURN(status=1)
          call WriteAttribute( hdf, 'lend', lend(region), status )
          IF_ERROR_RETURN(status=1)
        end if

        call WriteAttribute( hdf, 'xref'       , xref(region), status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'yref'       , yref(region), status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'zref'       , zref(region), status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'tref'       , tref(region), status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ntracet'    , ntracet     , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ntrace '    , ntrace      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nstd'       , nstd        , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'idate'      , idate       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'istart'     , istart      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ndyn'       , ndyn        , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nconv'      , nconv       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ndiag'      , ndiag       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nchem'      , nchem       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nsrce'      , nsrce       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nread'      , nread       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nwrite'     , nwrite      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ninst'      , ninst       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ncheck'     , ncheck      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ndiff'      , ndiff       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'itaui'      , itaui       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'itaue'      , itaue       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'itaut'      , itaut       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'itau0'      , itau0       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'idatei'     , idatei      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'idatee'     , idatee      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'idatet'     , idatet      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'idate0'     , idate0      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'icalendo'   , icalendo    , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'iyear0'     , iyear0      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'julday0'    , julday0     , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ndiagp1'    , ndiagp1     , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ndiagp2'    , ndiagp2     , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nstep'      , nstep       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'cpu0'       , cpu0        , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'cpu1'       , cpu1        , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'ra'         , ra          , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'fscale'     , fscale      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'names'      , names       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'areag'      , areag       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'czeta'      , czeta       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'czetak'     , czetak      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'xlabel'     , xlabel      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'istd'       , istd        , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'newyr'      , newyr       , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'newmonth'   , newmonth    , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'newday'     , newday      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'newsrun'    , newsrun     , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'cdebug'     , cdebug      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'limits'     , limits      , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'at'         , at          , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'bt'         , bt          , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'adv_scheme' , adv_scheme  , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'nsplitsteps', nsplitsteps , status )
        IF_ERROR_RETURN(status=1)
        call WriteAttribute( hdf, 'splitorder' , splitorder  , status )
        IF_ERROR_RETURN(status=1)
        
        msg_file = msg
        call WriteAttribute( hdf, 'msg', msg_file, status )
        IF_ERROR_RETURN(status=1)

        m => mass_dat(region)%m_t
        rm => mass_dat(region)%rm_t
        rxm => mass_dat(region)%rxm_t
        rym => mass_dat(region)%rym_t
        rzm => mass_dat(region)%rzm_t
#ifdef secmom
        rxxm => mass_dat(region)%rxxm_t
        rxym => mass_dat(region)%rxym_t
        rxzm => mass_dat(region)%rxzm_t
        ryym => mass_dat(region)%ryym_t
        ryzm => mass_dat(region)%ryzm_t
        rzzm => mass_dat(region)%rzzm_t
#endif

        imr = im(region) ; jmr = jm(region) ; lmr = lm(region)

        which_par=previous_par(region)

        if ( which_par /= 'tracer' ) call escape_tm('savehdf: Data should be parallel over tracers')

        call Init( sds, hdf, 'm', shape(m(1:imr,1:jmr,1:lmr)), 'real(4)', status )
        call SetDim(sds,0,'LON'//region_name(region),'deg',(/(xbeg(region)+i*dx/xref(region),i=0,imr-1)/),status)
        call SetDim(sds,1,'LAT'//region_name(region),'deg',(/(ybeg(region)+i*dy/yref(region),i=0,jmr-1)/),status)
        call SetDim(sds,2,'HYBRID','number',(/(i,i=1,lmr)/),status)
        call WriteData(sds,m(1:imr,1:jmr,1:lmr),status)
        call Done(sds,status)
        IF_ERROR_RETURN(status=1)

        call Init( sds, hdf, 'rm', (/imr,jmr,lmr,ntracetloc/), 'real(4)', status )
        call SetDim(sds,0,'LON'//region_name(region),'deg',(/(xbeg(region)+i*dx/xref(region),i=0,imr-1)/),status)
        call SetDim(sds,1,'LAT'//region_name(region),'deg',(/(ybeg(region)+i*dy/yref(region),i=0,jmr-1)/),status)
        call SetDim(sds,2,'HYBRID','number',(/(i,i=1,lmr)/),status)
        call SetDim(sds,3,'ntracet','None',(/(i,i=1,ntracetloc)/),status)
        call WriteData(sds,rm(1:imr,1:jmr,1:lmr,1:ntracetloc),status)
        call Done(sds,status)
        IF_ERROR_RETURN(status=1)

        call Init( sds, hdf, 'rxm', (/imr,jmr,lmr,ntracetloc/), 'real(4)', status )
        call SetDim(sds,0,'LON'//region_name(region),'deg',(/(xbeg(region)+i*dx/xref(region),i=0,im(region)-1)/),status)
        call SetDim(sds,1,'LAT'//region_name(region),'deg',(/(ybeg(region)+i*dy/yref(region),i=0,jm(region)-1)/),status)
        call SetDim(sds,2,'HYBRID','number',(/(i,i=1,lmr)/),status)
        call SetDim(sds,3,'ntracet','None',(/(i,i=1,ntracetloc)/),status)
        call WriteData(sds,rxm(1:imr,1:jmr,1:lmr,1:ntracetloc),status)
        call Done(sds,status)
        IF_ERROR_RETURN(status=1)

        call Init( sds, hdf, 'rym', (/imr,jmr,lmr,ntracetloc/), 'real(4)', status )
        call SetDim(sds,0,'LON'//region_name(region),'deg',(/(xbeg(region)+i*dx/xref(region),i=0,im(region)-1)/),status)
        call SetDim(sds,1,'LAT'//region_name(region),'deg',(/(ybeg(region)+i*dy/yref(region),i=0,jm(region)-1)/),status)
        call SetDim(sds,2,'HYBRID','number',(/(i,i=1,lmr)/),status)
        call SetDim(sds,3,'ntracet','None',(/(i,i=1,ntracetloc)/),status)
        call WriteData(sds,rym(1:imr,1:jmr,1:lmr,1:ntracetloc),status)
        call Done(sds,status)
        IF_ERROR_RETURN(status=1)

        call Init( sds, hdf, 'rzm', (/imr,jmr,lmr,ntracetloc/), 'real(4)', status )
        call SetDim(sds,0,'LON'//region_name(region),'deg',(/(xbeg(region)+i*dx/xref(region),i=0,im(region)-1)/),status)
        call SetDim(sds,1,'LAT'//region_name(region),'deg',(/(ybeg(region)+i*dy/yref(region),i=0,jm(region)-1)/),status)
        call SetDim(sds,2,'HYBRID','number',(/(i,i=1,lmr)/),status)
        call SetDim(sds,3,'ntracet','None',(/(i,i=1,ntracetloc)/),status)
        call WriteData(sds,rzm(1:imr,1:jmr,1:lmr,1:ntracetloc),status)
        call Done(sds,status)
        IF_ERROR_RETURN(status=1)


        call Done( hdf, status )
        IF_ERROR_RETURN(status=1)

        nullify(m)
        nullify(rm)
        nullify(rxm)
        nullify(rym)
        nullify(rzm)
#ifdef secmom
        nullify(rxxm)
        nullify(rxym)
        nullify(rxzm)
        nullify(ryym)
        nullify(ryzm)
        nullify(rzzm)
#endif
    end do  !regions...

    call Done( rcF, status )
    IF_NOTOK_RETURN(status=1)

    if(myid==root_t) call tstamp(kmain,itau,msg)

    if (cdebug) then
       call tstamp(kdebug,itau,'savehdf')
    end if
    
    call Par_Barrier

    ! ok
    status = 0

  end subroutine SaveHDF


end module io_save
