!-----------------------------------------------------------------------
!       purpose
!       -------
!       perform all calculations needed for TM5 test run  
!       with 1 ppb concentration
!
!       interface
!       ---------
!       call trace0
!       call trace1
!       call source
!       call chemie
!
!       method
!       ------
!       subroutine trace0 is called from start, and is used 
!                to input tracer specific values.
!       subroutine trace1 is called from start when istart = 2, and provides
!                a way to do some initialization, that has to be done only
!                at the beginning of a run.
!       subroutine source is called every nsrce seconds and provides 
!                the source/sink calculation.
!       subroutine chem  is called every nchem seconds and provides the
!                chemistry calculation on the tracer fields.
!       in zoom mode, recursive calling of the children is performed
!       
!-----------------------------------------------------------------------
!### macro's #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################

module sources_sinks

  use GO          , only : gol, goErr, goPr, goLabel
  use dims        , only : nregions
  use global_types, only : chem_data
  use global_data , only : outdir
  use emission    , only : Emis_Init, Emis_Done
  use chem_param
  use tracer_data,  only : Par_Check_Domain
  use toolbox,      only : escape_tm

  implicit none

  ! --- in/out --------------------------
  
  private
  
  public  ::  Sources_sinks_Init, Sources_sinks_Done
  public  ::  Trace0, Trace1, Trace_End, trace_after_read, start_co2
  public  ::  Source1, Source2
  public  ::  chem_dat
  
  ! --- const --------------------------------------

  character(len=*), parameter  ::  mname = 'sources_sinks'

  ! --- var ---------------------------
  
  type(chem_data), target  :: chem_dat(nregions)    ! tracers


contains


  ! ================================================================
  
  
  subroutine Sources_Sinks_Init( status )
  
    ! --- in/out --------------------------------
    
    integer, intent(out)           ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Sources_Sinks_Init'
    
    ! --- begin --------------------------------
    
    ! nothing to be init

    ! ok
    status = 0

    call Emis_Init(status)
    
  end subroutine Sources_Sinks_Init
  
  
  ! ***
  
  
  subroutine Sources_Sinks_Done( status )
  
    ! --- in/out --------------------------------
    
    integer, intent(out)           ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Sources_Sinks_Done'

    ! --- begin --------------------------------
    
    ! nothing to be done

    ! ok
    status = 0
    call Emis_Done(trim(outdir)//'/budget_global.hdf',status)
    
  end subroutine Sources_Sinks_Done
  
  
  
  ! ================================================================
  

  !-----------------------------------------------------------------------
  !      
  !       trace0
  !
  !
  !       purpose
  !       -------
  !       initialise data needed for chemistry, 
  !       e.g. read emissions and boundary conditions
  !       routine is calculated at start and at beginning each month...
  !
  !       interface
  !       ---------
  !       call trace0
  !
  !       method
  !       ------
  !       subroutine trace0 is called from module input, and provides a way
  !                to input tracer specific values.
  !       
  !       external
  !       --------
  !           calc_sm
  !           ini_zone
  !           ini_zoneg
  !           call_diagbud
  !           call_diagbudg
  !  
  !       reference
  !       ---------
  !       see above
  !---------------------------------------------------------------------   

  subroutine trace0( status )

    use dims,        only : im, jm, lm
    use chem_param,  only : ntracet, ntrace_chem

    use dims,          only : newmonth, idate, mlen, dxy11, nlat180, newsrun
    use dims,          only : sec_day,sec_month,sec_year
    use geometry,      only : calc_dxy
    use datetime,      only : calc_sm
#ifdef with_budgets
    use budget_fine,   only : diagbud,  ini_zone
    use budget_global, only : diagbudg, ini_zoneg
#endif
    use ParTools     , only : lmloc

    ! --- in/out --------------------------

    integer, intent(out)        ::  status

    ! --- local --------------------------

    integer  :: xref_, yref_, zref_, tref_
    integer  :: region
    integer  :: i, j, ii, n, nmonth
    integer  :: io, sfstart, ifail, sfend
    real     :: year2s, month2s, year2month
    integer  :: iyear, nxw, nyw, nzw, nmw

    ! --- begin ---------------------------

    write (gol,'("trace0: newmonth ",l1)') newmonth; call goPr
    write (gol,'("trace0: start month ",i2)') idate(2); call goPr

    ! calculate some conversion factors related to time...

    call calc_sm(mlen,sec_day,sec_month,sec_year)

    if(newsrun) then
        !
        ! calculate the area of the gridboxes on 1x1 degree
        !
        call calc_dxy(dxy11,nlat180)

#ifdef with_budgets
        call ini_zone    ! the budget of region = nregions...
        call ini_zoneg   ! the global budget

        do n=1,nregions
           call diagbud (n,1)               ! initialize budget files fine
           call diagbudg(n,1)               ! initialize budget files global
        end do
#endif

       ! allocate arrays for collecting data if necessary:
       do n = 1, nregions
         if ( .not. associated(chem_dat(n)%rm_k) ) then
           allocate( chem_dat(n)%rm_k(im(n),jm(n),lmloc, &
                                    ntracet+1:ntracet+ntrace_chem) )
         end if
       end do

    endif ! newsrun

    write (gol,'("trace0: end month ",i2)') idate(2); call goPr

    ! ok
    status = 0

  end subroutine trace0

  subroutine trace1
    !----------------------------------------------------------------------- 
    !
    !       purpose
    !       -------
    !       this subroutine determines the initial values for tracer mass and
    !       and its slopes.  it is called from input when istart = 2.
    !
    !       interface
    !       ---------
    !       call trace0
    !
    !       method
    !       ------
    !         
    !       reference
    !       ---------
    !       see above
    !
    !----------------------------------------------------------------------- 

    use dims,        only : im, jm, lm, adv_scheme
    use toolbox,     only : print_totalmass, escape_tm
    use global_data, only : mass_dat
    use zoom_tools,  only : update_parent
    use chem_param,  only : fscale, ntracet, ntrace, ntrace_chem
#ifdef MPI
    use mpi_const,   only : myid, lmar, ntracet_ar
#endif
    use ParTools   , only : ntracetloc, lmloc

    implicit none

    ! local
    real,dimension(:,:,:,:),pointer   :: rm, rxm, rym, rzm
    real,dimension(:,:,:),  pointer   :: m
    real,dimension(:,:,:,:),pointer   :: rmc
    integer                           :: region, n, offsetl, l, offsetn

    ! start

    if ( ntracetloc == 0 ) return !WP! only PE's with nonzero ntracetloc

    do region=1,nregions
       call Par_Check_Domain(region,'n','tracer') !WP! data must be on tracers for trace1

       m => mass_dat(region)%m_t
       rm => mass_dat(region)%rm_t
       rxm => mass_dat(region)%rxm_t
       rym => mass_dat(region)%rym_t
       rzm => mass_dat(region)%rzm_t
      do n=1,ntracetloc
       rm(1:im(region),1:jm(region),1:lm(region),n)=&
                 350.e-6*m(1:im(region),1:jm(region),1:lm(region))*xmco2/xmair
      enddo
       if(adv_scheme.eq.'slope') then
          rxm(1:im(region),1:jm(region),1:lm(region),:) = 0.0
          rym(1:im(region),1:jm(region),1:lm(region),:) = 0.0
          rzm(1:im(region),1:jm(region),1:lm(region),:) = 0.0
       endif
    enddo
    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)

#ifdef MPI
    write (gol,*) 'trace1: rm initialized at mixing ratio :', (myid+1)*1e-9; call goPr
#else
    write (gol,*) 'trace1: rm initialized at mixing ratio :', 1e-9; call goPr
#endif

    do region = nregions,2,-1
       call update_parent(region)
    end do
    call print_totalmass(1)

  end subroutine trace1

  subroutine start_co2
  use io_hdf
  use toolbox,     only : escape_tm, coarsen_emission
  use global_data, only : mass_dat
  use global_types,only : d3_data
  use zoom_tools,  only : update_parent
  use ParTools,    only : ntracetloc
  use Dims,        only : im,jm,lm,nregions, adv_scheme, nlon360,nlat180
  use emission_common,only : EmisInputDir
  implicit none
  !______________local______________________________
  real,dimension(:,:,:),pointer   ::m
  real,dimension(:,:,:,:),pointer   ::rm,rxm,rym,rzm
  real,dimension(nlon360,nlat180,lm(1))                :: field3d
  integer                                              :: io, sfstart, ifail, sds_id
  type(d3_data),dimension(nregions),target             :: co2_init
  integer,parameter                                    :: avg_field=1
  integer                                              :: region,l,imr,jmr,n, status


  call goLabel(mname//'/start_co2')

  if(ntracetloc == 0) return
  do region=1,nregions
    imr=im(region)
    jmr=jm(region)
    allocate(co2_init(region)%d3(imr,jmr,lm(1)))
  enddo

  io = sfstart(trim(EmisInputDir)//'co2_init.hdf', DFACC_READ)
  if ( io == FAIL ) then 
      write (gol,'("Failed to open co2 init file ...")'); call goErr
      write (gol,'("in program")'); call goErr; status=1
      call escape_tm('stop:')
  endif

  call io_read3d_32g(io,nlon360,nlat180,lm(1),field3d,'co2',ifail)

  if ( ifail /= 0 ) then
      write (gol,'("Failed to read co2 init field ...")'); call goErr
      write (gol,'("in program")'); call goErr; status=1; return
      call escape_tm('stop:')
  endif

  call coarsen_emission('co2_init',nlon360,nlat180,lm(1),field3d,co2_init,avg_field) ! not needed

  write (gol,*) 'Global avg CO2 in start-up field: ',sum(co2_init(1)%d3)/(im(1)*jm(1)*lm(1)) ; call goPr
  do region=1,nregions
      call Par_Check_Domain(region,'n','tracer') !WP! data must be on tracers for trace1
      m => mass_dat(region)%m_t
      rm => mass_dat(region)%rm_t
      rxm => mass_dat(region)%rxm_t
      rym => mass_dat(region)%rym_t
      rzm => mass_dat(region)%rzm_t
      do n=1,ntracetloc

         ! There is a bug with TM5 using the reduced grid.  Apparently
         ! negative concentrations are not permitted in this scenario.
         ! Thus for the forward simulation, we start each component
         ! with the background CO2 field, on which sinks can act.  To
         ! get the component concentration, the bg field must be
         ! subtracted.

         rm(1:im(region),1:jm(region),1:lm(region),n) = &
              co2_init(region)%d3*1.e-6*m(1:im(region),1:jm(region),1:lm(region))*xmco2/xmair

      enddo
      if(adv_scheme.eq.'slope') then
         rxm(:,:,:,:) = 0.0
         rym(:,:,:,:) = 0.0
         rzm(:,:,:,:) = 0.0
      endif
      nullify(m)
      nullify(rm)
      nullify(rxm)
      nullify(rym)
      nullify(rzm)
  enddo

  do region=1,nregions
    deallocate(co2_init(region)%d3)
  enddo
  
  do region = nregions,2,-1
     call update_parent(region)
  enddo

  end subroutine start_co2

  !--------------------------------------------------------------------- 
  ! 
  ! subroutine that is called after reading new fields 
  ! (clouds, surface winds, etc.)
  ! In this routine, 'chemistry' fields that depend on these data are calculated
  ! Routine is called from tracer.f90 (after reading data)
  ! 
  !--------------------------------------------------------------------- 

  subroutine trace_after_read
  
    ! --- begin ---------------------------------

    ! nothing to be done

  end subroutine trace_after_read

  subroutine source1(region)
    !--------------------------------------------------------------------- 
    ! 
    !       source1
    !
    !
    !       purpose
    !       -------
    !       this subroutine changes the tracer mass and its slopes by chemical sources
    !      
    !       interface
    !       ---------
    !       call source1
    !
    !       method
    !       ------
    !  
    !       
    !
    !       reference
    !       --------
    !       see above      
    !
    !---------------------------------------------------------------------
    use emission,      only: emission_apply
    implicit none
    !___________________IO_____________________
    integer, intent(in)        :: region
    !___________________local__________________

    call emission_apply(region)   ! note dust emissions are ported to sedimentation routine
  end subroutine source1


  subroutine source2(region)
    ! ---------------------------------------------------------------------
    !
    !       source2
    !
    !
    !       purpose
    !       -------
    !       apply boundary conditions to selected tracers
    !      
    !       interface
    !       ---------
    !       call source2
    !
    !       method
    !       ------
    !  
    !       
    !       external
    !       --------
    !       none
    !
    !       reference
    !       --------
    !       see above      
    !                  
    !
    !---------------------------------------------------------------------
    integer, intent(in) :: region

  end subroutine source2


  !
  ! deallocate the memory and do everything needed to finalise chemistry
  !

  subroutine trace_end( msg, file_name, status )
  
    use dims,          only : region_name, im, jm, lm
    use dims         , only : idate
    use chem_param,    only : ntracet, ntrace_chem
    use global_data,   only : mass_dat
    use file_hdf     , only : THdfFile, Init, Done
    use io_hdf,        only : io_write
#ifdef with_budgets
    use budget_fine,   only : diagbud
    use budget_global, only : diagbudg
    use budget_global, only : init_mass
    use budget_global, only : sum_chemistry, sum_update, apply_budget_global
    use budget_global, only : sum_advection, sum_deposition, sum_stratosphere
#endif
    use emission_data, only: sum_emission
#ifdef MPI
    use mpi_const, only : my_real, mpi_sum, com_lev
    use mpi_const, only : ntracetloc, lmloc, tracer_active, ierr
#endif
    use partools , only : myid, root_t, root_k
    use partools , only : Par_Gather_Tracer_k, Par_Barrier

    ! --- in/out ------------------------------------
    
    character(len=*), intent(in)  ::  file_name
    character(len=*), intent(in)  ::  msg
    integer, intent(out)          ::  status
    
    ! --- local -------------------------------------

    character(len=40)        :: frmt
    integer                  :: region, n, imr, jmr, lmr
    real                     :: mass, sum_chemistry_all, sum_wet_all
    real,dimension(:,:,:,:), pointer      :: rmc
    real,dimension(:,:,:,:), allocatable  :: fieldglob4d
    
    character(len=256)  ::  fname
    type(THdfFile)      ::  hdf

    ! --- begin --------------------------------------

#ifdef with_budgets
    if (apply_budget_global) then

      if ( myid == root_t ) then 
        write (gol,'("--------------------------------------------------------")'); call goPr
        write (gol,'("trace_end: Budget of tracer 1 in all regions (kg) :     ")'); call goPr
        write (gol,'("--------------------------------------------------------")'); call goPr
      end if

      do region =1,nregions 

        call diagbud(region,3)    !add depdry
        call diagbudg(region,3)    !add depdry
        call diagbud(region,2)    !write output
        call diagbudg(region,2)    !write output, free memory...

        mass = sum(mass_dat(region)%rm_t(1:im(region),1:jm(region),1:lm(region),1))

        call get_format(mass,frmt)

        if ( myid == root_t ) then 
          if (apply_budget_global) then
            write (gol,'("Region             ",i3)') region; call goPr
            write (gol,frmt) 'Initial mass   :            ',init_mass(region); call goPr
            write (gol,frmt) 'mass emitted   :            ',sum_emission(region); call goPr
            write (gol,frmt) 'chemistry      :            ',sum_chemistry(region); call goPr
            write (gol,frmt) 'update_p       :            ',sum_update(region); call goPr
            write (gol,frmt) 'advection      :            ',sum_advection(region); call goPr
            write (gol,frmt) 'deposition     :            ',sum_deposition(region); call goPr
            write (gol,frmt) 'stratosphere   :            ',sum_stratosphere(region); call goPr
            write (gol,frmt) 'Final mass     :            ',mass; call goPr
            write (gol,frmt) 'Budget (error) :            ', &
                 init_mass(region)  + sum_emission(region) &
                 +sum_advection(region) + sum_chemistry(region) &
                 +sum_update(region)  &
                 +sum_deposition(region) + sum_stratosphere(region) - mass; call goPr
            write (gol,'("--------------------------------------------------------")'); call goPr
          end if
        end if

      end do
      
    endif  !apply_budget_global
#endif

    do region=1,nregions
       deallocate(chem_dat(region)%rm_k)
    end do
    
    ! ok
    status = 0

  end subroutine trace_end


  ! ***

  
  subroutine get_format(mm,frmt)

    real,intent(in)              :: mm
    character(len=*),intent(out) :: frmt

    integer :: is

    if ( mm > 0.0 ) then
      is = int(alog10(mm))
    else
      is = 0
    end if

    select case(is)
      case(1:10)
        write (frmt,'(a9,i1,a1)') '(a30,f14.', 10-is, ')'
      case default
        frmt =  '(a30,1pe12.5)'
    end select

  end subroutine get_format


end module sources_sinks
